
import chainer.functions as F

from .gmm_layer import GMMLayer

class FVLayer(GMMLayer):

	def encode(self, x, use_mask=False, eps=1e-6):
		gamma = self.soft_assignment(x)
		_x, *params = self._expand_params(x)

		_gamma = self.xp.expand_dims(gamma.array, axis=2)
		_mu, _sig, _w = [p.array for p in params]

		"""
			If the GMM component is degenerate and has a null prior, then it
			must have null posterior as well. Hence it is safe to skip it.
			In practice, we skip over it even if the prior is very small;
			if by any chance a feature is assigned to such a mode, then
			its weight would be very high due to the division by
			priors[i_cl] below.

		"""
		# mask out all gammas, that are < eps
		eps_mask = (_gamma >= eps).astype(self.xp.float32)
		_gamma *= eps_mask

		# mask out all weights, that are < eps
		eps_mask = (_w >= eps).astype(self.xp.float32)
		_w = _w * eps_mask

		### Here the computations begin
		_std = F.sqrt(_sig)
		_x_mu_sig = (_x - _mu) / _std

		# TODO: use here the mask!
		mask = self.get_mask(x, use_mask)
		selected = self.xp.zeros(_x_mu_sig.shape, dtype=_x_mu_sig.dtype)
		selected[mask] = 1

		G_mu = _gamma * _x_mu_sig * selected
		G_sig = _gamma * (_x_mu_sig**2 - 1) * selected

		G_mu = F.sum(G_mu, axis=1) / selected.sum(axis=1)
		G_sig = F.sum(G_sig, axis=1) / selected.sum(axis=1)

		_w = self.xp.broadcast_to(self.w, G_mu.shape)
		G_mu /= self.xp.sqrt(_w)
		G_sig /= self.xp.sqrt(2 * _w)

		# 2 * (n, in_size, n_components) -> (n, 2, in_size, n_components)
		res = F.stack([G_mu, G_sig], axis=1)
		# (n, 2, in_size, n_components) -> (n, 2, n_components, in_size)
		res = res.transpose(0, 1, 3, 2)
		# res = F.stack([G_mu], axis=-1)
		# (n, 2, n_components, in_size) -> (n, 2*in_size*n_components)
		res = F.reshape(res, (x.shape[0], -1))
		return res

	def __call__(self, x, use_mask=False):
		_ = super(FVLayer, self).__call__(x, use_mask)
		return self.encode(x, use_mask)



from chainer.function import Function
from chainer import cuda

from .base import GammaMixin

class Gamma(Function, GammaMixin):

	def forward(self, inputs):
		self.y = self._gamma(inputs)
		return self.y,

	def backward(self, inputs, gys):
		xp = cuda.get_array_module(*inputs)

		g_gamma = self._gamma_grad(inputs)

		y = xp.broadcast_to(xp.expand_dims(self.y, axis=2), self._x_mu.shape)
		gy = xp.broadcast_to(xp.expand_dims(gys[0], axis=2), self._x_mu.shape)

		gx = (gy * g_gamma * y).sum(axis=3)
		return gx, None, None, None

def gamma(x, mu, sig, w, eps=None):
	return Gamma(eps=eps)(x, mu, sig, w)



import numpy as np

from unittest import TestCase

import chainer
import chainer.functions as F

from chainer_addons.links.fisher_encoding import FVLayer
from cyvlfeat.fisher import fisher
from cyvlfeat.gmm import gmm, cygmm

from .base import BaseFVEncodingTest

class FVLayerTest(BaseFVEncodingTest):

	def _new_layer(self, *args, **kwargs):
		return super(FVLayerTest, self)._new_layer(layer_cls=FVLayer)

	def test_output_shape(self):
		layer = self._new_layer()

		with chainer.using_config("train", False):
			output = layer(self.X)

		output_shape = (self.n, 2 * self.n_components * self.in_size)
		self.assertEqual(output.shape, output_shape,
			"Output shape was not correct!")

	def test_output(self):
		layer = self._new_layer()

		mean, var, w = (layer.mu, layer.sig, layer.w)

		with chainer.using_config("train", False):
			output = layer(self.X).array

		x = self.X.array.astype(np.float32)

		# we need to convert the array in Fortran order, but still remain the dimensions,
		# since the python API awaits <dimensions>x<components> arrays,
		# but they are indexed with "i_cl*dimension + dim" in the C-Code

		params = (
			mean.ravel(order="F").reshape(mean.shape, order="C").copy(),
			var.ravel(order="F").reshape(var.shape, order="C").copy(),
			w.copy(),
		)
		ref = [fisher(_x.T, *params,
					normalized=False,
					square_root=False,
					improved=False,
					fast=False,
					verbose=False,
			) for _x in x]

		ref = np.stack(ref)

		self.assertEqual(output.shape, ref.shape,
			"Output shape was not equal to ref shape!")

		self.assertClose(output, ref,
			"Output was not similar to reference")


	def test_cygmm(self):
		layer = self._new_layer()
		params = mean, var, w = (layer.mu, layer.sig, layer.w)

		x = self.X.array.astype(np.float32)
		gamma = layer.soft_assignment(x).array
		log_proba, _ = layer.log_proba(x, weighted=True)
		log_proba = log_proba.array

		for i, _x in enumerate(x):
			cy_mean, cy_var, cy_w, LL, cy_gamma = cygmm.cy_gmm(
				_x,
				self.n_components,
				0, # max_iterations
				"custom".encode("utf8"), # init_mode
				1, # n_repitions
				0, # verbose
				covariance_bound=None,

				init_means=mean.T.copy(),
				init_covars=var.T.copy(),
				init_priors=w.copy(),
			)

			self.assertClose(gamma[i], cy_gamma,
				f"[{i}] Soft assignment was not similar to reference (vlfeat)")

			self.assertClose(float(log_proba[i].sum()), LL,
				f"[{i}] Log-likelihood was not similar to reference (vlfeat)")



import numpy as np

import abc
import chainer

from unittest import TestCase

class BaseFVEncodingTest(abc.ABC, TestCase):

	def __init__(self, *args, **kwargs):
		super(BaseFVEncodingTest, self).__init__(*args, **kwargs)
		self.xp = np
		self.atol = 1e-4
		self.rtol = 1e-3

	def assertClose(self, arr0, arr1, msg):
		self.assertTrue(np.allclose(arr0, arr1, rtol=self.rtol, atol=self.atol),
			f"{msg}:\n{arr0}\n!=\n{arr1}")

	def setUp(self):
		self.n, self.t, self.in_size = 8, 4, 128
		self.n_components = 2
		self.alpha = 0.9

		self.seed = None
		self.dtype = np.float32

		self.rnd = self.xp.random.RandomState(self.seed)

		X = self.rnd.randn(self.n, self.t, self.in_size).astype(self.dtype)
		self.X = chainer.Variable(X)

		self.init_mu = self.rnd.randn(self.in_size, self.n_components).astype(self.dtype)
		self.init_sig = self.rnd.rand(self.in_size, self.n_components).astype(self.dtype)

		# [0 .. 1] -> [.5 .. 2]
		self.init_sig *= 1.5
		self.init_sig += .5

	@abc.abstractmethod
	def _new_layer(self, layer_cls, init_mu=None, init_sig=None):
		return layer_cls(
			self.in_size,
			self.n_components,
			use_sk_learn_em=True,
			init_mu=self.init_mu if init_mu is None else init_mu,
			init_sig=self.init_sig if init_sig is None else init_sig,
			alpha=self.alpha)



import numpy as np

from unittest import TestCase

import chainer

from chainer_addons.links.fisher_encoding import GMMLayer
from scipy.stats import multivariate_normal as mvn

from .base import BaseFVEncodingTest

class GMMLayerTest(BaseFVEncodingTest):

	def _new_layer(self, *args, **kwargs):
		return super(GMMLayerTest, self)._new_layer(layer_cls=GMMLayer)

	def test_output(self):
		layer = self._new_layer()

		res = layer(self.X)

		self.assertIs(res, self.X,
			"Output should be the input!")

	def test_update(self):
		layer = self._new_layer()
		params = (layer.mu, layer.sig, layer.w)

		params0 = [np.copy(p) for p in params]
		with chainer.using_config("train", False):
			y0 = layer(self.X)
		params1 = [np.copy(p) for p in params]

		for p0, p1 in zip(params0, params1):
			self.assertTrue(np.all(p0 == p1),
				"Params should not be updated when not training!")


		params0 = [np.copy(p) for p in params]
		with chainer.using_config("train", True):
			y0 = layer(self.X)
		params1 = [np.copy(p) for p in params]

		for p0, p1 in zip(params0, params1):
			self.assertTrue(np.all(p0 != p1),
				"Params should be updated when training!")

	def test_assignment_shape(self):
		layer = self._new_layer()

		gamma = layer.soft_assignment(self.X).array
		correct_shape = (self.n, self.t, self.n_components)
		self.assertEqual(gamma.shape, correct_shape,
			"Shape of the soft assignment is not correct!")

	def test_assignment_sum(self):
		layer = self._new_layer()

		gamma = layer.soft_assignment(self.X).array
		gamma_sum = gamma.sum(axis=-1)
		self.assertClose(gamma_sum, 1,
			"Sum of the soft assignment should be always equal to 1, but was")

	def test_assignment(self):
		layer = self._new_layer()
		correct_shape = (self.n, self.t, self.n_components)
		gamma = layer.soft_assignment(self.X).array

		gmm = layer.as_sklearn_gmm()
		x = self.X.reshape(-1, self.in_size).array
		ref_gamma = gmm.predict_proba(x).reshape(correct_shape)

		self.assertClose(gamma, ref_gamma,
			"Soft assignment is not similar to reference value from sklearn")


	def test_log_assignment(self):
		layer = self._new_layer()
		correct_shape = (self.n, self.t, self.n_components)
		log_gamma = layer.log_soft_assignment(self.X).array

		gmm = layer.as_sklearn_gmm()
		x = self.X.reshape(-1, self.in_size).array
		_, log_ref_gamma = gmm._estimate_log_prob_resp(x)
		log_ref_gamma = log_ref_gamma.reshape(correct_shape)

		self.assertClose(log_gamma, log_ref_gamma,
			"Log soft assignment is not similar to reference value from sklearn")

	def test_log_proba(self):

		layer = self._new_layer()

		params = mean, var, _ = (layer.mu, layer.sig, layer.w)
		mean, var = mean.T, var.T

		log_ps2, _ = layer.log_proba(self.X, weighted=False)
		ps2, _ = layer.proba(self.X, weighted=False)


		for i, x in enumerate(self.X.reshape(-1, self.in_size).array):
			n, t = np.unravel_index(i, (self.n, self.t))

			for component in range(self.n_components):
				log_p1 = mvn.logpdf(x, mean[component], var[component]).astype(self.dtype)
				log_p2 = log_ps2[n, t, component].array

				self.assertClose(log_p1, log_p2,
					f"{[n,t,component]}: Log-Likelihood was not the same")

				p1 = mvn.pdf(x, mean[component], var[component]).astype(self.dtype)
				p2 = ps2[n, t, component].array

				self.assertClose(p1, p2,
					f"{[n,t,component]}: Likelihood was not the same")



"""
Script for plotting loss and accuracy from chainer logs

usage:
	> python -m chainer_addon.utils.plot <logfile> [-k <key1> <key2> ... <keyn> ]
"""

import sys, argparse

parser = argparse.ArgumentParser(description = "Plot loss and accuracy from chainer logs")
parser.add_argument("logfile", type=str, help="chainer log file to parse")
parser.add_argument("--key","-k", type=str, default=["main/accuracy", "main/loss"], nargs="+", help="key of the loss and accuracy values")
parser.add_argument("--val_key", default="validation", type=str, help="key of the validation runs")
parser.add_argument("--fit", action="store_true", help="plot fitted graphs")
parser.add_argument("--track_changes", action="store_true", help="track changes in the log file and update the graphs accordingly")
parser.add_argument("--clip", type=int, default=0, help="clip values above to the given value (do not clip if 0)")
parser.add_argument("--all_in_one", "-aio", action="store_true", help="plot all graphs in one")

from os.path import join, basename, dirname
from scipy.optimize import curve_fit
import simplejson as json, numpy as np
import matplotlib.pyplot as plt
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler, FileModifiedEvent
from contextlib import contextmanager
from collections import defaultdict
import time

def func(x, a, b, c):
	return a * np.exp(-b * x) + c

def non_zero(idxs, values):
	return list(zip(*[(idx, val) for idx, val in zip(idxs, values) if val != 0]))



def update_graphs(args, data, validation_data, graphs):
	fig, ax = plt.figure(0), None
	fig.canvas.set_window_title(args.logfile)
	num_figs = len(data)
	if args.all_in_one:
		rows = cols = 1
	else:
		rows = np.ceil(np.sqrt(num_figs))
		cols = np.ceil(num_figs / rows)

	def extract_data(data):
		X, Y = map(np.array, zip(*data))
		if args.clip > 0:
			Y = np.minimum(Y, args.clip)
		return X, Y

	def set_line_data(line, X, Y, label):
		if (Y == 0).all(): return
		line.set_xdata(X)
		line.set_ydata(Y)
		line.set_label(label)

	for i, (key, values) in enumerate(data.items(), 1):
		if ax is None or not args.all_in_one:
			ax = fig.add_subplot(rows, cols, i)
			ax.set_title(key if not args.all_in_one else "graphs")

		assert None not in [fig, ax]

		X, Y = extract_data(values)
		if key in graphs:
			line = graphs[key]["line"]
			set_line_data(line, X, Y, key)

			if validation_data[key]:
				X, Y = extract_data(validation_data[key])
				line = graphs[key]["val_line"]
				set_line_data(line, X, Y, key + " val")

			ax = graphs[key]["axes"]
			ax.relim()
			ax.autoscale_view()
		else:
			graphs[key]["fignum"] = i
			graphs[key]["figure"] = fig
			graphs[key]["axes"] = ax
			if (Y != 0).any():
				graphs[key]["line"] = ax.plot(X, Y, label=key)[0]

			if args.fit:
				coefs = curve_fit(func, X, Y)[0]
				ax.plot(X, [func(x, *coefs) for x in X], "r--")

			if validation_data[key]:
				X, Y = extract_data(validation_data[key])
				graphs[key]["val_line"] = ax.plot(X, Y, label=key + " val")[0]

		ax.legend()
		ax.grid(True)


def read_logs(args, log_file):
	data = {key: [] for key in args.key}
	validation_data = {key: [] for key in args.key}
	for log in log_file:
		for key in args.key:
			data[key].append([int(log.get("epoch")), log.get(key, 0)])
			if key.startswith(args.val_key): continue
			val_key = "{}/{}".format(args.val_key, key)
			if val_key in log:
				validation_data[key].append([int(log.get("epoch")), log.get(val_key, 0)])

	return data, validation_data



@contextmanager
def track_change(args, callback):
	if not args.track_changes: yield; return
	path, fname = dirname(args.logfile), basename(args.logfile)

	class event_handler(FileSystemEventHandler):
		current_file = None
		def on_modified(self, event):
			if not isinstance(event, FileModifiedEvent): return
			if self.current_file == event.src_path: return
			if fname in event.src_path:
				print("Updating Graphs")
				callback()
				self.current_file = event.src_path

	obs = Observer()
	obs.schedule(event_handler(), path, recursive=False)
	obs.start()

	try:
		yield obs
	finally:
		obs.stop()
	obs.join(10)
	print()

def main(args):
	graphs = defaultdict(dict)
	def draw():
		log_file = json.load(open(args.logfile))#[:200]
		data, val_data = read_logs(args, log_file)
		update_graphs(args, data, val_data, graphs=graphs)
		plt.draw()

	with track_change(args, draw) as obs:
		if not args.track_changes:
			log_file = json.load(open(args.logfile))#[:200]
			data, val_data = read_logs(args, log_file)
			update_graphs(args, data, val_data, graphs=graphs)
			plt.show()
			return
		else:
			plt.ion()
			draw()

		try:
			while True:
				plt.pause(0.05)
				if not obs.is_alive(): break
		except KeyboardInterrupt:
			pass


main(parser.parse_args())





from .learning_rate import *
from .sacred_reporter import SacredReport
from .input_observer import InputObserver
from .alternate_trainable import AlternateTrainable
from .switch_trainables import SwitchTrainables
from .warm_up import WarmUp



from chainer.function import Function
from chainer import cuda
import numpy as np

class SignedSquareRoot(Function):

	def forward(self, inputs):
		xp = cuda.get_array_module(*inputs)
		x, = inputs
		return xp.sign(x) * xp.sqrt(xp.abs(x)),

	def backward_cpu(self, inputs, gys):
		x, gy = inputs[0], gys[0]#.copy()
		gx = np.zeros_like(gy)

		mask = x!=0
		gx[mask] = gy[mask] / (2 * np.sqrt(np.abs(x[mask])))
		return gx,

	def backward_gpu(self, inputs, gys):

		x = inputs[0]
		gx = cuda.elementwise(
			in_params="B mask, T x, T gy",
			out_params="T gx",
			operation="""
				gx = mask ? 0 : gy / (2 * sqrt(abs(x)));
			""",
			name="signed_square_root_bwd")(x==0, x, gys[0])

		return gx,


def signed_square_root(x):
	return SignedSquareRoot()(x)



#!/usr/bin/env python
if __name__ != '__main__': raise Exception("Do not import me!")

import sys
from os.path import abspath, dirname, join

addons_dir = abspath(join(abspath(dirname(__file__)), ".."))
sys.path.insert(0, addons_dir)

from tests.functions import *
from tests.links import *

import unittest

with chainer.using_config("train", False):
	unittest.main()




from .base import ContiniousLearningRate

class LinearLearningRate(ContiniousLearningRate):

	def new_lr(self, epoch):
		return self._lr - (self.factor * epoch)

	@property
	def factor(self):
		return (self._lr - self._target) / self._epochs



"""
here you can find some model definitions or modificated versions
of present models in chainer (eg. VGG19)
"""

from functools import partial

from .alexnet import AlexNet
from .vgg import VGG19Layers
from .resnet import ResnetLayers, Resnet35Layers
from .inception import InceptionV3
from .classifier import Classifier, RNNClassifier

from chainer_addons.utils.imgproc import generic_prepare, generic_tf_prepare
from chainer_addons.utils.imgproc import Size

from cvargparse.utils.enumerations import BaseChoiceType

class ModelType(BaseChoiceType):
	ResNet = ResnetLayers
	VGG19 = VGG19Layers
	Inception = InceptionV3

	Default = ResnetLayers

	def __call__(self, *args, **kwargs):
		"""
			Initializes new model instance
		"""
		model_cls = self.value
		model = model_cls(*args, **kwargs)
		return model

	@staticmethod
	def new(model_type, input_size, pooling, pooling_params={}, aux_logits=False):

		model_type = ModelType[model_type]
		kwargs = dict(
			pooling=pooling,
			pooling_params=pooling_params
		)

		if model_type == ModelType.Inception:
			kwargs["aux_logits"] = aux_logits

		if isinstance(input_size, (tuple, list)):
			input_size = Size(input_size)

		if isinstance(input_size, Size) or input_size > 0:
			model_type.value.meta.input_size = input_size

		return model_type(**kwargs)



class PrepareType(BaseChoiceType):
	MODEL = 0
	CUSTOM = 1
	TF = 2

	Default = CUSTOM

	def __call__(self, model):
		"""
			Initializes image preprocessing function
		"""

		if self == PrepareType.MODEL:
			return model.meta.prepare_func

		elif self == PrepareType.CUSTOM:
			return partial(generic_prepare,
				size=model.meta.input_size)

		elif self == PrepareType.TF:
			return generic_tf_prepare(
				size=model.meta.input_size,
				from_path=False)




import chainer
from chainer import link
from chainer import variable
from chainer import functions as F

import numpy as np

class GlobalAveragePooling(link.Link):
	output_dim=None

	# just ignore all Arguments
	def __init__(self, **kw):
		super(GlobalAveragePooling, self).__init__()

	def forward(self, x):
		n, channel, rows, cols = x.shape
		h = F.average_pooling_2d(x, (rows, cols), stride=1)
		h = F.reshape(h, (n, channel))
		return h



import chainer
import chainer.functions as F
import chainer.links as L
from chainercv.transforms import scale, resize

from chainer_addons.models.base import PretrainedModelMixin
from chainer_addons.links.pooling import PoolingType

from collections.abc import Iterable
from os.path import isfile

from .blocks import Inception1
from .blocks import Inception2
from .blocks import Inception3
from .blocks import Inception4
from .blocks import Inception5
from .blocks import AuxilaryClassifier
from .blocks import InceptionHead
from .link_mappings import chainer_to_keras, chainer_to_tf_ckpt

import numpy as np

def _assign(name, param, data):
	assert data.shape == param.shape, \
		"\"{}\" does not match the shape: {} != {}!".format(
			name, data.shape, param.shape)
	if isinstance(param, chainer.variable.Parameter):
		param.data[:] = data
	else:
		param[:] = data

def _assign_batch_norm(name, link, beta, avg_mean, avg_var):
	_assign(name, link.beta, beta)
	_assign(name, link.avg_mean, avg_mean)
	_assign(name, link.avg_var, avg_var)


class InceptionV3(PretrainedModelMixin, chainer.Chain):

	class meta:
		input_size = 299
		n_conv_maps = 2048
		feature_size = 2048
		mean = np.array([127.5, 127.5, 127.5], dtype=np.float32).reshape(3,1,1)

		classifier_layers = ["fc"]
		conv_map_layer = "mixed10"
		feature_layer = "pool"

		@staticmethod
		def prepare_func(x, size=None, swap_channels=True, keep_ratio=True):
			if size is None:
				size = InceptionV3.meta.input_size

			# [0 .. 255] -> [0 .. 1]
			x = x.astype(np.float32) / 255

			# HWC -> CHW channel order
			x = x.transpose(2,0,1)

			if keep_ratio:
				if isinstance(size, Iterable):
					size = min(size)
				# scale the smallest side to <size>
				x = scale(x, size=size)
			else:
				if isinstance(size, int):
					size = (size, size)

				# resize the image to  side to <size>x<size>
				x = resize(x, size)

			if swap_channels:
				x = x[::-1]
			return x

	def __init__(self, pretrained_model=None, aux_logits=False, *args, **kwargs):
		self.aux_logits = aux_logits
		super(InceptionV3, self).__init__(*args, **kwargs)

		if pretrained_model is not None and isfile(pretrained_model):
			self.load(pretrained_model, strict=True)

	def load(self, weights, *args, **kwargs):
		if weights.endswith(".h5"):
			self._load_from_keras(weights)
		elif weights.endswith(".ckpt.npz"):
			self._load_from_ckpt_weights(weights)
		else:
			return super(InceptionV3, self).load(weights, *args, **kwargs)

	def init_layers(self, n_classes):
		# input 3 x 299 x 299
		self.head = InceptionHead()
		# output 192 x 35 x 35


		pool_args = dict(pool_type=PoolingType.TF_AVG, ksize=3, stride=1, pad=1)
		# input 192 x 35 x 35
		self.mixed00 = Inception1(insize=192, sizes=[48, 64, 96], outputs=[64, 64, 96, 32], **pool_args)
		# input 256 x 35 x 35
		self.mixed01 = Inception1(insize=256, sizes=[48, 64, 96], outputs=[64, 64, 96, 64], **pool_args)
		# input 288 x 35 x 35
		self.mixed02 = Inception1(insize=288, sizes=[48, 64, 96], outputs=[64, 64, 96, 64], **pool_args)

		pool_args = dict(pool_type=PoolingType.MAX, ksize=3, stride=2, pad=0)
		# input 288 x 35 x 35
		self.mixed03 = Inception2(288, sizes=[64, 96], outputs=[384, 96], **pool_args)
		# output 288 x 17 x 17

		# input 768 x 17 x 17
		pool_args = dict(pool_type=PoolingType.TF_AVG, ksize=3, stride=1, pad=1)
		# input 768 x 17 x 17
		self.mixed04 = Inception3(768, sizes=[128] * 6, outputs=[192] * 4, **pool_args)
		# input 768 x 17 x 17
		self.mixed05 = Inception3(768, sizes=[160] * 6, outputs=[192] * 4, **pool_args)
		# input 768 x 17 x 17
		self.mixed06 = Inception3(768, sizes=[160] * 6, outputs=[192] * 4, **pool_args)
		# input 768 x 17 x 17
		self.mixed07 = Inception3(768, sizes=[192] * 6, outputs=[192] * 4, **pool_args)

		self.aux = AuxilaryClassifier(n_classes) if self.aux_logits else F.identity

		pool_args = dict(pool_type=PoolingType.MAX, ksize=3, stride=2, pad=0)
		# input 768 x 17 x 17
		self.mixed08 = Inception4(768, sizes=[192] * 4, outputs=[320, 192], **pool_args)
		# output 768 x 8 x 8

		# input 1280 x 8 x 8
		pool_args = dict(pool_type=PoolingType.TF_AVG, ksize=3, stride=1, pad=1)
		# input 1280 x 8 x 8
		self.mixed09 = Inception5(1280, sizes=[384, 448, 384], outputs=[320, 384, 384, 192], **pool_args)
		# input 2048 x 8 x 8
		self.mixed10 = Inception5(2048, sizes=[384, 448, 384], outputs=[320, 384, 384, 192], **pool_args)

		# input 2048 x 8 x 8
		# global average pooling
		# output 2048 x 1 x 1
		self.fc = L.Linear(2048, n_classes)

	@property
	def _links(self):
		names = ["mixed{:02d}".format(i) for i in range(11)]
		body = [(name, [getattr(self, name)]) for name in names]
		return [
			("head", [self.head]),
		] + body + [
			("pool", [self.pool]),
			("fc", [self.fc]),
		]

	def extract(self, x):
		x = self.head(x)
		x = self.mixed00(x)
		x = self.mixed01(x)
		x = self.mixed02(x)
		x = self.mixed03(x)
		x = self.mixed04(x)
		x = self.mixed05(x)
		x = self.mixed06(x)
		x = self.mixed07(x)
		x = self.mixed08(x)
		x = self.mixed09(x)
		x = self.mixed10(x)

		return self.pool(x)


	def __call__(self, x, layer_name='fc'):
		aux_logit = None
		for key, funcs in self.functions.items():
			for func in funcs:
				x = func(x)

				if chainer.config.train and self.aux_logits and func == self.mixed07:
					aux_logit = self.aux(x)

			if key == layer_name:
				return x, aux_logit


	def loss(self, pred, gt, loss_func=F.softmax_cross_entropy, alpha=0.4):
		if isinstance(pred, tuple):
			pred0, aux_pred = pred
		else:
			pred0, aux_pred = pred, None

		loss = loss_func(pred0, gt)
		if aux_pred is None:
			return loss
		else:
			aux_loss = loss_func(aux_pred, gt)
			return (1-alpha) * loss + alpha * aux_loss

	def accuracy(self, pred, gt):
		if isinstance(pred, tuple):
			pred0, aux_pred = pred
		else:
			pred0, aux_pred = pred, None
		return F.accuracy(pred0, gt)

	def _load_from_ckpt_weights(self, weights):
		content = np.load(weights)
		for name, link in self.namedlinks(skipself=True):
			if name not in chainer_to_tf_ckpt:
				continue

			ckpt_key = chainer_to_tf_ckpt[name]

			if isinstance(link, L.Convolution2D):
				W = content["{}/weights".format(ckpt_key)]
				W = W.transpose(3,2,0,1)
				_assign(name, link.W, W)

			elif isinstance(link, L.BatchNormalization):
				beta = content["{}/beta".format(ckpt_key)]
				avg_mean = content["{}/moving_mean".format(ckpt_key)]
				avg_var = content["{}/moving_variance".format(ckpt_key)]
				_assign_batch_norm(name, link, beta, avg_mean, avg_var)

			elif isinstance(link, L.Linear):
				W = content["{}/weights".format(ckpt_key)]
				W = W.transpose(3,2,0,1).squeeze()
				b = content["{}/biases".format(ckpt_key)]
				_assign(name, link.W, W)
				_assign(name, link.b, b)

			else:
				raise ValueError("Unkown link type: {}!".format(type(link)))

	def _load_from_keras(self, weights):
		import h5py
		with h5py.File(weights, "r") as f:
			if "model_weights" in f:
				f = f["model_weights"]
			for name, link in self.namedlinks(skipself=True):
				if name not in chainer_to_keras: continue
				keras_key = chainer_to_keras[name]

				if isinstance(link, L.Convolution2D):
					W = np.asarray(f[keras_key][keras_key + "/kernel:0"])
					W = W.transpose(3,2,0,1)

					_assign(name, link.W, W)

				elif isinstance(link, L.Linear):
					W = np.asarray(f[keras_key][keras_key + "/kernel:0"])
					b = np.asarray(f[keras_key][keras_key + "/bias:0"])

					_assign(name, link.W, W.transpose(1,0))
					_assign(name, link.b, b)

				elif isinstance(link, L.BatchNormalization):
					beta = np.asarray(f[keras_key][keras_key + "/beta:0"])
					avg_mean = np.asarray(f[keras_key][keras_key + "/moving_mean:0"])
					avg_var = np.asarray(f[keras_key][keras_key + "/moving_variance:0"])

					_assign_batch_norm(name, link, beta, avg_mean, avg_var)
				else:
					raise ValueError("Unkown link type: {}!".format(type(link)))




import numpy as np, chainer
import os, re
from os.path import isdir, join
from chainer.training import extension
from collections import defaultdict
from chainer import training

class InputObserver(extension.Extension):
	def __init__(self, output, match_regex=None, gather_trigger=(1, "iteration"), write_out_trigger=(1, "epoch")):
		super(InputObserver, self).__init__()
		self.output = output
		self.match_regex = match_regex
		self.gather_trigger = training.trigger.get_trigger(gather_trigger)
		self.write_out_trigger = training.trigger.get_trigger(write_out_trigger)

		if not isdir(output):
			os.makedirs(output)

		self.inputs = defaultdict(list)

	def __call__(self, trainer):
		self.gather(trainer)
		self.write_out(trainer)

	def matches(self, key):
		return self.match_regex is None or \
			re.search(self.match_regex, key) is not None

	def gather(self, trainer):
		if not self.gather_trigger(trainer): return
		updater = trainer.updater
		model = updater.get_optimizer('main').target
		for child in model.children():
			for name, link in child.namedlinks(skipself=True):
				if not hasattr(link, "inputs"): continue
				key = "{}/{}".format(child.name, name)
				if not self.matches(key): continue
				inputs = chainer.cuda.to_cpu(link.inputs[0].data)
				self.inputs[key].append(inputs.astype(np.float16))

	def write_out(self, trainer):
		if not self.write_out_trigger(trainer): return
		updater = trainer.updater
		epoch = updater.epoch
		iteration = updater.iteration
		fname = join(self.output, "{}_{}".format(epoch, iteration))
		inputs = {key: np.vstack(values) for key, values in self.inputs.items()}
		self.inputs.clear()
		np.savez(fname, **inputs)




from chainer.training import StandardUpdater

class MiniBatchUpdater(StandardUpdater):
	"""
		The iterator outputs batches in mini-batch sizes. This updater
		cummulates the gradients of these mini-batches until the
		batch_size is reached. Then a parameter update is performed
	"""
	def __init__(self, update_size=32, *args, **kwargs):
		super(MiniBatchUpdater, self).__init__(*args, **kwargs)
		self.update_size = update_size
		self.iteration_counter = 0

	def update_core(self):
		optimizer = self.get_optimizer('main')
		loss_func = self.loss_func or optimizer.target
		it = self.get_iterator('main')

		batch = it.next()
		data = self.converter(batch, self.device)

		use_cleargrads = getattr(optimizer, '_use_cleargrads', True)
		if use_cleargrads and self.iteration_counter == 0:
			optimizer.target.cleargrads()

		self.iteration_counter += it.batch_size
		loss = loss_func(*data)
		loss.backward()

		if self.iteration_counter >= self.update_size:
			self.iteration_counter = 0
			optimizer.update()



import abc
import numpy as np

from functools import partial

from chainer import cuda

class GammaMixin(abc.ABC):
	TWO_SQ_PI = np.sqrt(2 * np.pi)

	def __init__(self, eps):
		super(GammaMixin, self).__init__()
		self.eps = eps

	# @staticmethod
	# def expand_params(params, shape, xp=np):
	# 	bcast = partial(xp.broadcast_to, shape=shape)
	# 	return map(bcast, params)

	def _log_proba(self, x, mu, sigma, xp=np):
		assert x.ndim == 3, \
			"Number of dimensions of the input is not 3: {}!".format(x.ndim)

		n, t, in_size = x.shape
		_, n_components = mu.shape
		shape  = (n, t, in_size, n_components)

		_x = xp.broadcast_to(xp.expand_dims(x, -1), shape)
		_mu, _sig = [xp.broadcast_to(p, shape) for p in [mu, sigma]]

		_dist = (_x - _mu) / _sig
		return -xp.sum(_dist**2, axis=2) / 2, (_mu, _sig)

	def _proba(self, x, mu, sigma, w,
		weighted=True, return_params=False, xp=np):

		_log_proba, (_mu, _sig) = self._log_proba(x, mu, sigma, xp=xp)
		_proba = xp.exp(_log_proba)

		_w = xp.broadcast_to(w, _proba.shape)
		if weighted:
			_proba = xp.sum(_proba * _w, axis=-1)

		if return_params:
			return _proba, (_mu, _sig, _w)
		else:
			return _proba

	def _gamma(self, inputs, eps=None):
		xp = cuda.get_array_module(*inputs)

		x, mu, sigma, w = inputs
		n, t, in_size = x.shape
		_, n_components = mu.shape

		_log_proba, (_mu, _sig) = self._log_proba(x, mu, sigma, xp=xp)

		shape = (n, t, in_size, n_components)
		_x = xp.broadcast_to(xp.expand_dims(x, -1), shape)
		self._x_mu = _x - _mu
		self._sig = _sig

		_w = xp.broadcast_to(w, _log_proba.shape)
		self._log_wu = _log_proba + xp.log(_w)
		if xp == np:
			from scipy.special import logsumexp
			_log_wu_sum = logsumexp(self._log_wu, axis=-1, keepdims=True)
		else:
			import pdb; pdb.set_trace()

		# _wu_sum = xp.sum(self._wu, axis=-1, keepdims=True) #+ eps
		return xp.exp(self._log_wu - _log_wu_sum)

	def _gamma_grad(self, inputs, eps=None):
		eps = self.eps if eps is None else eps
		xp = cuda.get_array_module(*inputs)
		x, mu, sigma, w = inputs

		n, t, in_size = x.shape
		_, n_components = mu.shape

		_log_wu = xp.expand_dims(self._log_wu, axis=2)
		_log_x_mu_sig_2 = xp.log(self._x_mu) - 2 * xp.log(self._sig)

		_log_x_mu_sig_2[np.isnan(_log_x_mu_sig_2)] = 0

		if xp == np:
			from scipy.special import logsumexp
			_log_wu_sum = logsumexp(self._log_wu, axis=-1, keepdims=True)
			_log_wu_sum = xp.expand_dims(_log_wu_sum, axis=-1)
			_log_wu_sum = xp.broadcast_to(_log_wu_sum, _log_x_mu_sig_2.shape)
			# import pdb; pdb.set_trace()
			_log_g0 = _log_x_mu_sig_2 + _log_wu_sum

			_log_wu = xp.expand_dims(self._log_wu, axis=2)
			_log_wu = xp.broadcast_to(_log_wu, _log_x_mu_sig_2.shape)
			_log_g1 = logsumexp(_log_wu + _log_x_mu_sig_2, axis=-1, keepdims=True)

			# import pdb; pdb.set_trace()
			# sum_kwargs = dict(axis=3, keepdims=True)
			# g0 = _log_x_mu_sig_2
			# g1 = xp.sum(_wu * _x_mu_sig_2, **sum_kwargs) \
			# 	/ xp.sum(_wu, **sum_kwargs)

			return xp.exp(_log_g1) - xp.exp(_log_g0)
			res = np.log(res) - log
		else:
			import pdb; pdb.set_trace()


		return g0 + g1



import chainer
from chainer.backends import cuda
from chainer import functions as F

from sklearn.utils import check_random_state

import abc
import numpy as np

_LOG_2PI = np.log(2 * np.pi)

def _kernel_e_step(X, means, cov, ws, xp=cuda.cupy):

	t, size = X.shape
	n_comp, size = means.shape
	log_det = xp.sum(xp.log(1. / xp.sqrt(cov)), axis=1)


	log_prob = xp.zeros((t, n_comp), dtype=xp.float32)
	log_prob_kernel = cuda.elementwise(
		name="gmm_log_prob",
		in_params="raw T X, T means, T cov, int32 t, int32 size, int32 n_comp",
		out_params="raw T log_prob",
		operation="""
			int ni = i / size; /* component idx */
			int fi = i % size; /* feature idx */
			int j = 0;
			int xi = 0;

			for ( /* part idx */ int ti = 0; ti < t; ti++) {
				int j = ti * n_comp + ni;
				int xi = ti * size + fi;

				/* printf( "%d, %d, %d -> %d, %d \\n", ti, ni, fi, j, xi); */
				atomicAdd( &log_prob[j], powf(X[xi] - means, 2) / cov);
			}
		"""
	)

	log_prob_kernel(X, means, cov, t, size, n_comp, log_prob)

	weighting_kernel = cuda.elementwise(
		name="gmm_weighted_prob",
		in_params="T log_prob, T ws, T log_det, int32 size",
		out_params="T weighted_log_prob",
		operation="weighted_log_prob = -0.5 * (size * LOG_2PI + log_prob) + log_det + log(ws);",
		preamble="#define LOG_2PI log(2 * 3.14159265359)",
	)
	log_prob = weighting_kernel(log_prob, ws, log_det, size)

	norm_kernel = cuda.elementwise(
		name="gmm_norm_log_prob",
		in_params="T log_prob, raw T norm, int32 n_comp",
		out_params="T norm_log_prob",
		operation="norm_log_prob = log_prob - norm[i/n_comp];",
	)

	log_prob_norm = F.logsumexp(log_prob, axis=1).array
	log_resp = norm_kernel(log_prob, log_prob_norm, n_comp)

	return xp.mean(log_prob_norm), log_resp

def _basic_e_step(X, means, cov, ws, xp=np):

	n_features = X.shape[1]

	log_det = xp.sum(xp.log(1. / xp.sqrt(cov)), axis=1)
	precisions = 1. / cov

	res0 = xp.sum((means ** 2 * precisions), 1)
	res1 = -2. * xp.dot(X, (means * precisions).T)
	res2 = xp.dot(X ** 2, precisions.T)

	log_prob = res0 + res1 + res2
	log_prob = -.5 * (n_features * _LOG_2PI + log_prob) + log_det

	weighted_log_prob = log_prob + xp.log(ws)

	log_prob_norm = F.logsumexp(weighted_log_prob, axis=1).array

	log_resp = weighted_log_prob - log_prob_norm[:, None]

	return xp.mean(log_prob_norm), log_resp


class GPUMixin(abc.ABC):

	def xp_from_array(self, X):
		if hasattr(X, "array"):
			_x = X.array
		else:
			_x = X

		return chainer.backend.get_array_module(_x)


	def _transform_X(self, X):
		X = X.reshape(-1, X.shape[-1])
		xp = self.xp_from_array(X)
		if hasattr(X, "array"):
			return X.array, xp
		else:
			return X, xp

	def _initialize_parameters(self, X, random_state):
		super(GPUMixin, self)._initialize_parameters(cuda.to_cpu(X), random_state)
		xp = self.xp_from_array(X)
		if xp != np:
			self.means_ = xp.array(self.means_, dtype=X.dtype)
			self.covariances_ = xp.array(self.covariances_, dtype=X.dtype)
			self.weights_ = xp.array(self.weights_, dtype=X.dtype)
			self.precisions_cholesky_ = xp.array(self.precisions_cholesky_, dtype=X.dtype)


	def fit(self, X, y=None):
		X, xp = self._transform_X(X)
		self._check_initial_parameters(X)

		random_state = check_random_state(self.random_state)
		attrs = ["means_", "covariances_", "weights_"]
		if not all([hasattr(self, attr) for attr in attrs]):
			self._initialize_parameters(X, random_state)

		for n_iter in range(1, self.max_iter + 1):
			log_prob_norm, log_resp = self._e_step(X, xp=xp)
			self._m_step(X, log_resp, xp=xp)

	def _e_step(self, X, xp=np, use_kernel=True):
		""" E step.
			Copied from sklearn/mixture/base.py
		"""
		if xp != np and use_kernel:
			e_step_impl = _kernel_e_step
		else:
			e_step_impl = _basic_e_step

		return e_step_impl(X, self.means_, self.covariances_, self.weights_, xp=xp)

	@abc.abstractmethod
	def _m_step(self, *args, **kwargs):
		raise NotImplementedError()

	def _gaussian_params(self, X, log_resp, xp):

		resp = xp.exp(log_resp)
		nk = resp.sum(axis=0) + 10 * xp.finfo(resp.dtype).eps
		means = xp.dot(resp.T, X) / nk[:, None]

		avg_X2 = xp.dot(resp.T, X ** 2) / nk[:, None]
		avg_means2 = means ** 2
		avg_X_means = means * xp.dot(resp.T, X) / nk[:, None]
		covariances = avg_X2 - 2 * avg_X_means + avg_means2

		covariances = xp.maximum(covariances, self.reg_covar)

		return nk, means, covariances

# testing the runtimes
if __name__ == '__main__':
	import time
	from functools import partial

	n_iter, warm_up = 10000, 200
	def bench(func, *args, **kwargs):
		global n_iter, warm_up

		for _ in range(warm_up):
			res = func(*args, **kwargs)

		t0 = time.time()

		for _ in range(n_iter):
			res = func(*args, **kwargs)

		t0 = time.time() - t0
		func_name = func.func.__name__ if isinstance(func, partial) else func.__name__
		print(f"{func_name} took {t0:.3f}s for {n_iter} runs")
		return res

	T, N_COMP, SIZE = 64, 10, 32
	dtype, xp = np.float32, cuda.cupy

	X = xp.random.randn(T, SIZE).astype(dtype)
	mu = xp.random.rand(N_COMP, SIZE).astype(dtype)
	sig = xp.random.rand(N_COMP, SIZE).astype(dtype) + 1
	ws = xp.ones(N_COMP).astype(dtype) / N_COMP

	logL0, log_gammas0 = bench(_kernel_e_step, X, mu, sig, ws, xp=xp)

	logL1, log_gammas1 = bench(_basic_e_step, X, mu, sig, ws, xp=xp)

	assert xp.allclose(log_gammas0, log_gammas1, atol=1e-5), \
		f"Were not equal: \n{log_gammas0} !=\n{log_gammas1}"

	assert xp.allclose(logL0, logL1, atol=1e-5), \
		f"Were not equal: \n{logL0} !=\n{logL1}\n"



import chainer
from chainer.functions.pooling.average_pooling_2d import AveragePooling2D
from chainer import cuda


class TFAveragePooling2D(AveragePooling2D):
	"""
		The only thing, that changes is the CuDNN poolgin option from
			CUDNN_POOLING_AVERAGE_COUNT_INCLUDE_PADDING
		to
			CUDNN_POOLING_AVERAGE_COUNT_EXCLUDE_PADDING

		Note: CuDNN is required for this Function
	"""

	def __init__(self, *args, **kwargs):
		super(TFAveragePooling2D, self).__init__(*args, **kwargs)


	def create_pool_desc(self):
		assert chainer.should_use_cudnn('>=auto'), \
			"This function works only with CuDNN!"
		return cuda.cudnn.create_pooling_descriptor(
			(self.kh, self.kw), (self.sy, self.sx), (self.ph, self.pw),
			cuda.cuda.cudnn.CUDNN_POOLING_AVERAGE_COUNT_EXCLUDE_PADDING)


def tf_average_pooling_2d(x, ksize, stride=None, pad=0):
	return TFAveragePooling2D(ksize, stride, pad, cover_all=False).apply((x,))[0]



import numpy as np
import chainer
import chainer.links as L
import chainer.functions as F

from chainer.serializers import npz
from chainer.links.model.vision.resnet import prepare, BuildingBlock
from functools import partial

from .base import PretrainedModelMixin

class ResnetMixin(PretrainedModelMixin):

	@property
	def _links(self):
		links = [
				('conv1', [self.conv1, self.bn1, F.relu]),
				('pool1', [partial(F.max_pooling_2d, ksize=3, stride=2)]),
				('res2', [self.res2]),
				('res3', [self.res3]),
				('res4', [self.res4]),
				('res5', [self.res5]),
				('pool5', [self.pool])]
		if hasattr(self, "fc6"):
			links +=[
				('fc6', [self.fc6]),
				('prob', [F.softmax]),
			]

		return links


class ResnetLayers(ResnetMixin, L.ResNet50Layers):

	class meta:
		classifier_layers = ["fc6"]
		conv_map_layer = "res5"
		feature_layer = "pool5"
		feature_size = 2048
		n_conv_maps = 2048
		input_size = 448
		mean = np.array([103.063,  115.903,  123.152], dtype=np.float32).reshape(3,1,1)

		@staticmethod
		def prepare_func(x, size=None, *args, **kwargs):
			if size is None:
				size = ResnetLayers.meta.input_size
			elif isinstance(size, int):
				size = (size, size)

			return prepare(x, size)

	def init_layers(self, n_classes, **kwargs):
		pass


class Resnet35Layers(ResnetMixin, chainer.Chain):

	class meta:
		classifier_layers = ["fc6"]
		conv_map_layer = "res5"
		feature_layer = "pool5"
		feature_size = 2048
		n_conv_maps = 2048
		input_size = 448
		mean = np.array([103.063,  115.903,  123.152], dtype=np.float32).reshape(3,1,1)

		prepare_func = prepare


	def init_layers(self, n_classes, **kwargs):
		self.conv1 = L.Convolution2D(3, 64, 7, 2, 3, **kwargs)
		self.bn1 = L.BatchNormalization(64)
		self.res2 = BuildingBlock(2, 64, 64, 256, 1, **kwargs)
		self.res3 = BuildingBlock(3, 256, 128, 512, 2, **kwargs)
		self.res4 = BuildingBlock(3, 512, 256, 1024, 2, **kwargs)
		self.res5 = BuildingBlock(3, 1024, 512, 2048, 2, **kwargs)
		self.fc6 = L.Linear(2048, n_classes)

	def __call__(self, x,  layer_name=None):
		layer_name = layer_name or self.meta.classifier_layers[-1]
		for key, funcs in self.functions.items():
			for func in funcs:
				x = func(x)
			if key == layer_name:
				return x
